#include<string.h>
#include<stdlib.h>
#include"argon.h"

#include"GL_helper.h"
#include"calibration.h"

#include<unistd.h>
#include<time.h>

#include<GL/glut.h>

#include <HD/hd.h>
#include <HDU/hdu.h>
#include <HDU/hduError.h>
#include <HDU/hduVector.h>

t_state *state;
t_parameters pars;

static HHD ghHD = HD_INVALID_HANDLE;

static HDSchedulerHandle hUpdateDeviceCallback;

static HDdouble workspacemodel[16];
static double kCursorScreenSize = 20.0;
static double gCursorScale = 1.0;

double forceScale=1;
double maxForce=1;

float atomRadius=0.1;

int activeAtom=-1;


void displayFunction(void);

void updateWorkspace();

typedef struct
{
	int button1;
	int button2;
    hduVector3Dd position;
    hduVector3Dd force;
    HDdouble transform[16];
    hduVector3Dd userForce;
    HDboolean isUserForceActive;
} HapticDisplayState;


HapticDisplayState deviceState;

/*******************************************************************************
 This handler will get called when the application is exiting. This is our
 opportunity to cleanly shutdown the HD API.
*******************************************************************************/
void exitHandler(void)
{
    hdStopScheduler();
    hdUnschedule(hUpdateDeviceCallback);

	printf("exitting\n");

    if (ghHD != HD_INVALID_HANDLE)
    {
        hdDisableDevice(ghHD);
        ghHD = HD_INVALID_HANDLE;
    }

}

HDCallbackCode HDCALLBACK copyHapticDisplayState(void *pUserData)
{
	memcpy(pUserData, &deviceState, sizeof(deviceState));

    return HD_CALLBACK_DONE;
}


HDCallbackCode HDCALLBACK setHapticUserForce(void *pUserData)
{
	memcpy(deviceState.userForce, pUserData, sizeof(hduVector3Dd));
	deviceState.isUserForceActive=1;

    return HD_CALLBACK_DONE;
}

HDCallbackCode HDCALLBACK resetHapticUserForce(void *pUserData)
{
	deviceState.isUserForceActive=0;

    return HD_CALLBACK_DONE;
}

HDCallbackCode HDCALLBACK updateDeviceCallback(void *data)
{
	/* Poczatek ramki haptycznej */
	hdBeginFrame(ghHD);
	int buttons=0;

	hdGetIntegerv(HD_CURRENT_BUTTONS, &buttons);

	deviceState.button1=(buttons & HD_DEVICE_BUTTON_1) ? 1 : 0;
	deviceState.button2=(buttons & HD_DEVICE_BUTTON_2) ? 1 : 0;

    hdGetDoublev(HD_CURRENT_POSITION, deviceState.position);
    hdGetDoublev(HD_CURRENT_TRANSFORM, deviceState.transform);
    hdGetDoublev(HD_CURRENT_FORCE, deviceState.force);

	hduVector3Dd force;
	memset(force, 0, sizeof(hduVector3Dd));

	if(deviceState.isUserForceActive) {
		memcpy(force, deviceState.userForce, sizeof(hduVector3Dd));

		double modForce=hduVecMagnitude(force);

		if(modForce>maxForce) {
			hduVecScaleInPlace(force, maxForce/modForce); 
		}
	}

	hdSetDoublev(HD_CURRENT_FORCE, force);

	/* Koniec ramki. */
	hdEndFrame(ghHD);

	/* Sprawdzenie, czy nie wystapil blad. */
	HDErrorInfo error;
	if (HD_DEVICE_ERROR(error = hdGetError())) {
		hduPrintError(stderr, &error, "Error\n");

		if (hduIsSchedulerError(&error)) {
			return HD_CALLBACK_DONE;
		}
	}

	/* Ten callback ma byc wywolany rowniez w nastepnej iteracji schedulera. */
	return HD_CALLBACK_CONTINUE;
}

void initHD()
{
    HDErrorInfo error;
    ghHD = hdInitDevice(HD_DEFAULT_DEVICE);
    if (HD_DEVICE_ERROR(error = hdGetError())) 
    {
        hduPrintError(stderr, &error, "Failed to initialize haptic device");
        fprintf(stderr, "\nPress any key to quit.\n");
        getchar();
        exit(-1);
    }

    hdEnable(HD_FORCE_OUTPUT);

    hdStartScheduler();
    if (HD_DEVICE_ERROR(error = hdGetError()))
    {
        hduPrintError(stderr, &error, "Failed to start the scheduler");
        exit(-1);
    }
	calibrate_haptic();

	printf("Press any key\n");
	getchar();

    hUpdateDeviceCallback = hdScheduleAsynchronous(updateDeviceCallback, 0, HD_MAX_SCHEDULER_PRIORITY);
}
/*******************************************************************************
  Draws a 3D cursor for the haptic device using the current local transform,
  the workspace to world transform and the screen coordinate scale.
 *******************************************************************************/
void drawCursor(const HapticDisplayState *pState)
{
	static const double kCursorRadius = 0.5;
	static const double kCursorHeight = 1.5;
	static const int kCursorTess = 15;

	static GLuint gCursorDisplayList = 0;

	GLUquadricObj *qobj = 0;

	glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_LIGHTING_BIT);
	glPushMatrix();

	if (!gCursorDisplayList)
	{
		gCursorDisplayList = glGenLists(1);
		glNewList(gCursorDisplayList, GL_COMPILE);
		qobj = gluNewQuadric();

		gluCylinder(qobj, 0.0, kCursorRadius, kCursorHeight,
				kCursorTess, kCursorTess);
		glTranslated(0.0, 0.0, kCursorHeight);
		gluCylinder(qobj, kCursorRadius, 0.0, kCursorHeight / 5.0,
				kCursorTess, kCursorTess);

		gluDeleteQuadric(qobj);
		glEndList();
	}

	/* Apply the workspace view transform. */
	glMultMatrixd(workspacemodel);

	/* Apply the local transform of the haptic device. */
	glMultMatrixd(pState->transform);

	/* Apply the local cursor scale factor. */
	glScaled(gCursorScale, gCursorScale, gCursorScale);

	glEnable(GL_COLOR_MATERIAL);
	glColor3f(0.0, 0.5, 1.0);

	glCallList(gCursorDisplayList);

	glPopMatrix(); 
	glPopAttrib();
}

/*******************************************************************************
  Uses the current OpenGL viewing transforms to initialize a transform for the
  haptic device workspace so that it's properly mapped to world coordinates.
 *******************************************************************************/
void updateWorkspace()
{
	HDdouble screenTworkspace;
	GLdouble modelview[16];
	GLdouble projection[16];
    GLint viewport[4];

    glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    /* Compute the transform for going from device coordinates to world
       coordinates based on the current viewing transforms. */
    hduMapWorkspaceModel(modelview, projection, workspacemodel);

    /* Compute the scale needed to display the cursor of a particular size
       in screen coordinates. */
    screenTworkspace = hduScreenToWorkspaceScale(
        modelview, projection, viewport, workspacemodel);

    gCursorScale = kCursorScreenSize * screenTworkspace;
}

void getModelCoord(t_coord *res, hduVector3Dd vec)
{
	t_coord pos={vec[0], vec[1], vec[2], 0};

	res->x=pos.x*workspacemodel[0] + pos.y*workspacemodel[4] + pos.z*workspacemodel[8]  + workspacemodel[12];
	res->y=pos.x*workspacemodel[1] + pos.y*workspacemodel[5] + pos.z*workspacemodel[9]  + workspacemodel[13];
	res->z=pos.x*workspacemodel[2] + pos.y*workspacemodel[6] + pos.z*workspacemodel[10] + workspacemodel[14];
}

void getHapticCoord(hduVector3Dd res, t_coord vec)
{
	t_coord pos=vec;

	pos.x-=workspacemodel[12];
	pos.y-=workspacemodel[13];
	pos.z-=workspacemodel[14];

	double det=0;

	det+=workspacemodel[0] * workspacemodel[5] * workspacemodel[10];
	det+=workspacemodel[1] * workspacemodel[6] * workspacemodel[8];
	det+=workspacemodel[2] * workspacemodel[4] * workspacemodel[9];

	det-=workspacemodel[0] * workspacemodel[6] * workspacemodel[9];
	det-=workspacemodel[1] * workspacemodel[4] * workspacemodel[10];
	det-=workspacemodel[2] * workspacemodel[5] * workspacemodel[8];


	res[0]=pos.x*workspacemodel[0] + pos.y*workspacemodel[1] + pos.z*workspacemodel[2];
	res[1]=pos.x*workspacemodel[4] + pos.y*workspacemodel[5] + pos.z*workspacemodel[6];
	res[2]=pos.x*workspacemodel[8] + pos.y*workspacemodel[9] + pos.z*workspacemodel[10];

	double scale=pow(det, -2.0/3.0);

	res[0]*=scale;
	res[1]*=scale;
	res[2]*=scale;
}

/*******************************************************************************
 Graphics main loop function.
*******************************************************************************/
void displayFunction(void)
{
    // Setup model transformations.
    glMatrixMode(GL_MODELVIEW); 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glPushMatrix();

    setupGraphicsState();

    // Draw the fixed sphere.
    float atomColor[4] = {.2, .8, .8, .8};
    float selAtomColor[4] = {.8, .2, .2, .8};
    float actAtomColor[4] = {.2, .8, .2, .8};
    float lockAtomColor[4] = {.2, .5, .5, .8};


    HapticDisplayState stateHD;

    /* Obtain a thread-safe copy of the current haptic display state. */
    hdScheduleSynchronous(copyHapticDisplayState, &stateHD, HD_DEFAULT_SCHEDULER_PRIORITY);
	


	if(activeAtom>=0) {
		for(int i=0; i<state->N; i++) {
			if(i==activeAtom) {
				drawSphere((double*)&(state->r[i]), actAtomColor, atomRadius*1.1);
			} else {
				if(state->locked[i]) {
					drawSphere((double*)&(state->r[i]), lockAtomColor, atomRadius);
				} else {
					drawSphere((double*)&(state->r[i]), atomColor, atomRadius);
				}

			}
		}
	} else {
		t_coord cursor={0,0,0,0};
		getModelCoord(&cursor, (stateHD.position));
		for(int i=0; i<state->N; i++) {
			if(dist(state->r[i], cursor)<atomRadius) {
				drawSphere((double*)&(state->r[i]), selAtomColor, atomRadius*1.1);
			} else {
				if(state->locked[i]) {
					drawSphere((double*)&(state->r[i]), lockAtomColor, atomRadius);
				} else {
					drawSphere((double*)&(state->r[i]), atomColor, atomRadius);
				}
			}
		}
	}

	drawCursor(&stateHD);


	glPopMatrix();

	glutSwapBuffers();                      

	updateWorkspace();
}

void redrawTimer(int msec) {
	static int lastButton1=0;
	static int lastButton2=0;

	one_step(state, &pars);
	HapticDisplayState stateHD;

    /* Obtain a thread-safe copy of the current haptic display state. */
    hdScheduleSynchronous(copyHapticDisplayState, &stateHD, HD_DEFAULT_SCHEDULER_PRIORITY);
	
	t_coord cursor={0,0,0,0};
	getModelCoord(&cursor, (stateHD.position));

//	hduVector3Dd vec;
//	getHapticCoord(vec, cursor);
//	printf("%lf %lf %lf %lf %lf %lf\n", stateHD.position[0], stateHD.position[1], stateHD.position[2], vec[0], vec[1], vec[2]);
/*
	if(activeAtom>=0) {
		t_coord anchor={0,0,0,0};
		getModelCoord(&anchor, (stateHD.anchor));
		
	printf("%lf %lf %lf %lf %lf %lf\n", state->r[activeAtom].x, state->r[activeAtom].y, state->r[activeAtom].z, anchor.x, anchor.y, anchor.z);

	}
*/

	if(stateHD.button1==0 && stateHD.button2==0) {
		if(lastButton1)	hdScheduleSynchronous(resetHapticUserForce, 0, HD_DEFAULT_SCHEDULER_PRIORITY);

		activeAtom=-1;
	
	} else if(activeAtom==-1) {
		for(int i=0; i<state->N; i++) {
			if(dist(state->r[i], cursor)<atomRadius) {
				printf("activated: %d\n", i);
				activeAtom=i;
			}
		}
	}

	if(stateHD.button1) {
		if(activeAtom>=0) {
			if(lastButton1) {
				t_coord pos={0,0,0,0};
				getModelCoord(&pos, (stateHD.position));

				state->r[activeAtom]=pos;
				state->damping[activeAtom]=1;
			} 			t_coord force=scale(state->F[activeAtom], 0.1);

			hdScheduleSynchronous(setHapticUserForce, &(force), HD_DEFAULT_SCHEDULER_PRIORITY);
		}
	} else if(stateHD.button2) {
		if(activeAtom>=0) {
			if(!lastButton2) {
				state->locked[activeAtom]= state->locked[activeAtom]==1 ? 0 : 1;
			}
		}
	}

	lastButton1=stateHD.button1;
	lastButton2=stateHD.button2;

	glutTimerFunc(msec, redrawTimer, msec);
	glutPostRedisplay();
}

int main(int argc, char **argv)
{

	if(argc<1) {
		printf("Usage: %s <input file>\n", argv[0]);
		exit(-1);
	}

	srand(time(NULL)%getpid()+20);

	char *fname_in=argv[1];


	load_parameters(&pars, fname_in);

	atomRadius=pars.R/2;
		
	state=alloc_state(pars.N);

	init_state(state, &pars);
	compute_forces(state, &pars);

	for(int i=0; i<state->N-1; i++) {
		state->locked[i]=1;
	}

	initHD();
	
	initGlut(argc, argv);

    glutDisplayFunc(displayFunction); 

	glutTimerFunc(5, redrawTimer, 5);
    atexit(exitHandler);

	glutMainLoop();
}
